/*
 * linux/arch/arm/drivers/char/avr.c
 *
 * Copyright 2001 Applied Data Systems
 *
 */

#include <linux/config.h>

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/kbd_ll.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/kbd_kern.h>

#include <asm/irq.h>
#include <asm/hardware.h>
#include <linux/tqueue.h>
#include <linux/proc_fs.h>
#include "avrprg.h"

#define	DEBUG(args...)	printk(args)

#define	MAX_LOOP		30
static int	flash_size;
static int	flash_part;
static int	flash_family;

typedef	struct	t_AVR_CMD {
	unchar	byte1;
	unchar	byte2; 
	unchar	byte3;
	unchar	byte4; 
}	AVR_CMD;

static void send_SSP_msg(unchar *pBuf, int num)
{
	ushort	tmp;
	int		i;

	for (i=0;i<num;i++) {
		while ((Ser4SSSR & SSSR_TNF) == 0);
		tmp = pBuf[i];
		Ser4SSDR = (tmp << 8);
	}
	
	// Throw away Echo
	for (i=0;i<num;i++) {
		while ((Ser4SSSR & SSSR_RNE) == 0);
		tmp = Ser4SSDR;
	}
}

static unchar ReadSSPByte(void)
{
	if (Ser4SSSR & SSSR_ROR) {
		printk("%s() : Overrun\n", __FUNCTION__);
		return 0;
	}

	Ser4SSDR = 0x00;

	while ((Ser4SSSR & SSSR_RNE) == 0);

	return ((unchar) Ser4SSDR);
}

static ulong read_SSP_response(int num)
{
	int		i;
	ulong	ret;
	
	// discard leading 0x00 and command echo 0 (command group value)
	ReadSSPByte();
	ReadSSPByte();
	// discard command echo 1 (command code value)
	ReadSSPByte();

	// data from SMARTIO
	// It assumes LSB first.
	// NOTE:Some command uses MSB first order
	ret = 0;
	for (i=0;i<num;i++) {
		ret |= ReadSSPByte() << (8*i);
	}

	return ret;
}

static AVR_CMD ProgramEnable = { 0xAC, 0x53, 0x00, 0x00 };
static AVR_CMD ReadDevCode   = { 0x30, 0x00, 0x00, 0x00 };
static AVR_CMD ReadPrgFlash  = { 0x20, 0x00, 0x00, 0x00 };
static AVR_CMD WritePrgFlash = { 0x40, 0x00, 0x00, 0x00 };
static AVR_CMD ChipErase     = { 0xAC, 0x80, 0x00, 0x00 };
static AVR_CMD WriteLockBit  = { 0xAC, 0xF9, 0x00, 0x00 };
static AVR_CMD ReadLockBit   = { 0x58, 0x00, 0x00, 0x00 };

static int send_program_msg(unchar *pBuf, unchar *echo)
{
	ushort	tmp;
	int		i;

	for (i=0;i<4;i++) {
		while ((Ser4SSSR & SSSR_TNF) == 0);
		tmp = pBuf[i];
		Ser4SSDR = (tmp << 8);
	}
	
	// Throw away Echo
	for (i=0;i<4;i++) {
		while ((Ser4SSSR & SSSR_RNE) == 0);
		echo[i] = Ser4SSDR;
	}

	if ((pBuf[0] == echo[1]) && (pBuf[1] == echo[2])) {
		return 1;
	}
	else {
		return 0;
	}
}

static int avr_program_enable()
{
	unchar	echo[4];
	int		retry;

#if defined(CONFIG_SA1100_ADSBITSY) || defined(CONFIG_SA1100_ADSBITSYPLUS)
	GPDR |= GPIO_GPIO25;
	GPCR |= GPIO_GPIO25;
	mdelay(300);
#elif defined(CONFIG_SA1100_ADSAGC)
	GAFR &= ~GPIO_GPIO11;
	GPDR |= GPIO_GPIO11;
	GPCR |= GPIO_GPIO11;
	mdelay(300);
#else
	ADS_AVR_REG |= 0x01;
	mdelay(100);
	ADS_AVR_REG &= 0xFE;
	mdelay(100);
#endif

	retry = 0;
	while (!send_program_msg((unchar *)&ProgramEnable, echo)) {
		mdelay(20);
		if (retry++ > 32) break;
	}
	if (retry > 32) 
		return 0;
	else 
		return 1;
}

void avr_program_disable()
{
#if defined(CONFIG_SA1100_ADSBITSY) || defined(CONFIG_SA1100_ADSBITSYPLUS)
	GPDR |= GPIO_GPIO25;
	GPCR |= GPIO_GPIO25;
	mdelay(300);
	GPSR |= GPIO_GPIO25;
	mdelay(300);
#elif defined(CONFIG_SA1100_ADSAGC)
	GPDR |= GPIO_GPIO11;
	GPCR |= GPIO_GPIO11;
	mdelay(300);
	GPSR |= GPIO_GPIO11;
	mdelay(300);
#else
	// Reset SMARTIO
	ADS_AVR_REG &= 0xFE;
	mdelay(100);
	ADS_AVR_REG |= 0x01;
	mdelay(100);
#endif
}

static int avr_read_devcode(unchar *vendor, unchar *family, unchar *part)
{
	unchar	echo[4];

	ReadDevCode.byte3 = 0x00;
	if (send_program_msg((unchar *)&ReadDevCode, echo)) {
		if (echo[3] && (echo[3] != 0x1E)) {
			printk("AVR Vendor Error: ret = 0x%02x\n", echo[3]);
			return 0;
		}
		*vendor = echo[3];
	}
	else {
	        printk("send_program_msg 1 failed!\n");
		return 0;
	}

	ReadDevCode.byte3 = 0x01;
	if (send_program_msg((unchar *)&ReadDevCode, echo)) {
		if (echo[3] == 0x90) 
			flash_size = 1024;		// 1Kbyte
		else if (echo[3] == 0x91)
			flash_size = 2048;		// 2Kbyte
		else if (echo[3] == 0x92)
			flash_size = 4096;		// 4Kbyte
		else if (echo[3] == 0x93)
			flash_size = 8192;		// 8Kbyte
		else if (echo[3] == 0xFF) {
			printk("AVR Device Code Erased: ret = 0x%02x\n", echo[3]);
			return 0;
		}
		else if (echo[3] != 0x01) {
			printk("AVR Flash size Error: ret = 0x%02x\n", echo[3]);
			return 0;
		}
		*family = flash_family = echo[3];
	}
	else {
	        printk("send_program_msg 2 failed!\n");
		return 0;
	}

	ReadDevCode.byte3 = 0x02;
	if (send_program_msg((unchar *)&ReadDevCode, echo)) {
		*part = flash_part = echo[3];
		if ((*vendor == 0x00) && (*family == 0x01) && (*part == 0x02))
			printk("AVR is locked\n");
	}
	else {
	        printk("send_program_msg 3 failed!\n");
		return 0;
	}

	return 1;
}

static void set_SSP_highspeed()
{
	Ser4SSCR0 = 0x1007;
	Ser4SSSR = SSSR_ROR;
	Ser4SSCR1 = 0x0000;
	Ser4SSCR0 = 0x1087;

	mdelay(10);
}

static void set_SSP_lowspeed()
{
	Ser4SSCR0 = 0x4007;
	Ser4SSSR = SSSR_ROR;
	Ser4SSCR1 = 0x0000;
	Ser4SSCR0 = 0x4087;
	mdelay(10);
}

static void set_SSP_normalspeed()
{
	Ser4SSCR0 = 0xA707;
	Ser4SSSR = SSSR_ROR;
	Ser4SSCR1 = 0x0010;
	Ser4SSCR0 = 0xA787;
	mdelay(10);
}

void set_SSP_init(void)
{
#if !defined(CONFIG_SA1100_ADSBITSYPLUS) && !defined(CONFIG_SA1100_ADSAGC)
	GPDR |= (GPIO_GPIO10 | GPIO_GPIO12 | GPIO_GPIO13); 	// Output
	GPDR &= ~GPIO_GPIO11;

	// Alternative Function
	GAFR |= (GPIO_GPIO10 | GPIO_GPIO11 | GPIO_GPIO12 | GPIO_GPIO13);
#endif

#if defined(CONFIG_SA1100_ADSBITSY) || defined(CONFIG_SA1100_ADSBITSYPLUS)
	GPDR |= GPIO_GPIO25;
	GPCR |= GPIO_GPIO25;
	mdelay(300);
	GPSR |= GPIO_GPIO25;
	mdelay(300);
#elif defined(CONFIG_SA1100_ADSAGC)
	GAFR &= ~GPIO_GPIO11;
	GPDR |= GPIO_GPIO11;
	GPCR |= GPIO_GPIO11;
	mdelay(300);
	GPSR |= GPIO_GPIO11;
	mdelay(300);
#else
	// Reset SMARTIO
	ADS_AVR_REG &= 0xFE;
	mdelay(100);
	ADS_AVR_REG |= 0x01;
	mdelay(100);
#endif
}

static	AVR_CMD READ_DEVVER_CMD  = { 0x82, 0x05, 0x00, 0x00 };
static	AVR_CMD READ_DEVTYPE_CMD = { 0x82, 0x06, 0x00, 0x00 };
static	AVR_CMD READ_FWLEVEL_CMD = { 0x82, 0x07, 0x00, 0x00 };

static ushort read_sio_version(struct sio_ver *ptr)
{
	ushort			ret;

	// Read Device Version 
	send_SSP_msg((unchar *) &READ_DEVVER_CMD, 2);
	ret = read_SSP_response(1);
	ptr->DevVer = (uint)ret;
	// Read Device Type
	send_SSP_msg((unchar *) &READ_DEVTYPE_CMD, 2);
	ret = read_SSP_response(2);
	ptr->DevType = (uint)ret;
	// Read Firmware Level
	send_SSP_msg((unchar *) &READ_FWLEVEL_CMD, 2);
	ret = read_SSP_response(2);
	// swap MSB & LSB
	ret = ((ret & 0xFF) << 8) | ((ret & 0xFF00) >> 8);
	ptr->FwLevel = (uint)ret;

	return 0;
}

static int avr_open(struct inode *inode, struct file *file)
{
	unchar	vendor, family, part;
#if defined(CONFIG_SA1100_ADSAGC)
	ADS_INT_EN1 &= ~0x01; // disable ssp interrupts (to fix missing pull down resistor)
#endif

	set_SSP_highspeed();
	if (!avr_program_enable()) {
                printk("Unable to enable programming.\n");
		return -EIO;
	}

	if (!avr_read_devcode(&vendor, &family, &part)) {
                printk("Unable to read device code.\n");
		return -EIO;
	}

	avr_program_disable();
	set_SSP_normalspeed();

	return 0;
}


static int avr_release(struct inode *inode, struct file *file)
{
#if defined(CONFIG_SA1100_ADSAGC)
	ADS_INT_EN1 &= ~0x01; // disable ssp interrupts (to fix missing pull down resistor)
#endif
	avr_program_disable();

	set_SSP_normalspeed();

	return 0;
}

static ssize_t avr_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	*data = (unchar *)buf;
	unchar	upper, lower;
	unchar	echo[4];
	int		i, retry;

	if (minor != 0) return -ENXIO;

	if (count > flash_size) {
		printk("Requested size is greater than AVR Flash Size\n");
		return -EINVAL;
	}

	set_SSP_highspeed();

	if (!avr_program_enable())
		return -EIO;

	upper = lower = 0;
	for (i=0;i<count/2;i++) {
		ReadPrgFlash.byte1 = 0x20;			// Low Byte
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		retry = 0;
		while (!send_program_msg((unchar *)&ReadPrgFlash, echo)) {
			if (retry++ > MAX_LOOP) {
				avr_program_disable();
				printk("Device is busy..\n");
				return -EBUSY;
			}
		}
		*data++ = echo[3];

		ReadPrgFlash.byte1 = 0x28;			// High Byte
		retry = 0;
		while (!send_program_msg((unchar *)&ReadPrgFlash, echo)) {
			if (retry++ > MAX_LOOP) {
				avr_program_disable();
				printk("Device is busy..\n");
				return -EBUSY;
			}
		}
		*data++ = echo[3];

		if (lower == 0xFF) upper++;
		lower ++;
	}

	if (count%2) {
		ReadPrgFlash.byte1 = 0x20;			// Low Byte
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		retry = 0;
		while(!send_program_msg((unchar *)&ReadPrgFlash, echo)) {
			if (retry++ > MAX_LOOP) {
				avr_program_disable();
				printk("Device is busy.. 1\n");
				return -EBUSY;
			}
		}
		*data++ = echo[3];
	}
	avr_program_disable();

	return count;
}

static ssize_t avr_write_old(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	echo[4];
	unchar	upper, lower;
	unchar	*data = (unchar *)buf;
	int		i, retry;

	if (minor != 0) return -ENXIO;

	if (count > flash_size) {
		printk("Requested size is greater than AVR Flash Size\n");
		return -EINVAL;
	}
	if (!avr_program_enable())
		return -EIO;

	set_SSP_lowspeed();

	upper = lower = 0;
	for (i=0;i<count/2;i++) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = upper;
		WritePrgFlash.byte3 = lower;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		mdelay(5);
		ReadPrgFlash.byte1 = 0x20;
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		for (retry=0;retry<MAX_LOOP;retry++) {
			if (*data == 0xFF)  {
				mdelay(10);
				break;	
			}
			send_program_msg((unchar *)&ReadPrgFlash, echo);
			if (*data == echo[3])
				break;
			mdelay(5);
		}
		if (retry == MAX_LOOP) {
			avr_program_disable();
			printk("Device is busy..\n");
			return -EBUSY;
		}
		else
			data++;

		WritePrgFlash.byte1 = 0x48;			// High Byte
		WritePrgFlash.byte2 = upper;
		WritePrgFlash.byte3 = lower;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		mdelay(5);
		ReadPrgFlash.byte1 = 0x28;
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		for (retry=0;retry<MAX_LOOP;retry++) {
			if (*data == 0xFF)  {
				mdelay(10);
				break;	
			}
			send_program_msg((unchar *)&ReadPrgFlash, echo);
			if (*data == echo[3])
				break;
			mdelay(5);
		}
		if (retry == MAX_LOOP) {	
			avr_program_disable();
			printk("Device is busy..\n");
			return -EBUSY;
		}
		else
			data++;

		if (lower == 0xFF) upper++;
		lower ++;
	}

	if (count%2) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = upper;
		WritePrgFlash.byte3 = lower;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		mdelay(5);
		ReadPrgFlash.byte1 = 0x20;
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		for (retry=0;retry<MAX_LOOP;retry++) {
			if (*data == 0xFF)  {
				mdelay(10);
				break;	
			}
			send_program_msg((unchar *)&ReadPrgFlash, echo);
			if (*data == echo[3])
				break;
			mdelay(5);
		}
		if (retry == MAX_LOOP) {
			avr_program_disable();
			printk("Device is busy..\n");
			return -EBUSY;
		}
	}

	set_SSP_highspeed();
	avr_program_disable();

	return count;
}

static ssize_t avr_write_new(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	echo[4];
	unchar	upper, lower;
	unchar	*data = (unchar *)buf;
	int	i;
	int     page, page_count;

	if (minor != 0) return -ENXIO;

	if (count > flash_size) {
		printk("Requested size is greater than AVR Flash Size\n");
		return -EINVAL;
	}
	if (!avr_program_enable())
		return -EIO;

	set_SSP_lowspeed();

	upper = lower = 0;

	page_count = count / 64;

	for (page=0; page < page_count; page++) {

	  for (i=0; i<32; i++) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));

		data++;
		WritePrgFlash.byte1 = 0x48;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;
	  }

	  // Write page to flash
	  WritePrgFlash.byte1 = 0x4C;
	  WritePrgFlash.byte2 = ((page >> 3) & 0xF);
	  WritePrgFlash.byte3 = ((page & 0x7) << 5);
	  WritePrgFlash.byte4 = 0x00;
	  send_program_msg((unchar *)&WritePrgFlash, echo);
	  mdelay(5);
	}

	if (count % 64) {

	  for (i=0; i < ((count % 64) / 2); i++) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;

		WritePrgFlash.byte1 = 0x48;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;
	  }

	  if (count & 0x1) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;

	  }

	  // Write page to flash
	  page = page_count;
	  WritePrgFlash.byte1 = 0x4C;
	  WritePrgFlash.byte2 = ((page >> 3) & 0xF);
	  WritePrgFlash.byte3 = ((page & 0x7) << 5);
	  WritePrgFlash.byte4 = 0x00;
	  send_program_msg((unchar *)&WritePrgFlash, echo);
	  mdelay(5);
	}

	// high fuse
	mdelay(5);
	WritePrgFlash.byte1 = 0xAC;
	WritePrgFlash.byte2 = 0xA8;
	WritePrgFlash.byte3 = 0x00;
	WritePrgFlash.byte4 = 0x49;
	send_program_msg((unchar *)&WritePrgFlash, echo);

	// fuse
	mdelay(5);
	WritePrgFlash.byte1 = 0xAC;
	WritePrgFlash.byte2 = 0xA0;
	WritePrgFlash.byte3 = 0x00;
	WritePrgFlash.byte4 = 0xEF;
	send_program_msg((unchar *)&WritePrgFlash, echo);

	mdelay(100);

	set_SSP_highspeed();
	avr_program_disable();

	return count;
}

static ssize_t avr_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
        int result;
	if (flash_family == 0x93 && flash_part == 0x08) {
                // ATmega new part
	  result = avr_write_new(file, buf, count, ppos);
	}
	else {
	  result = avr_write_old(file, buf, count, ppos);
	}

	return result;
}

static int avr_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	vendor, family, part, echo[4];
	struct avr_ver	*DevVer;

	if (minor != 0) return -ENXIO;

	set_SSP_highspeed();

	switch (cmd) {
		case	READ_AVR_VER:
				set_SSP_normalspeed();
				read_sio_version((struct sio_ver *) arg);
				mdelay(100);				// delay for stable operation
				return sizeof(struct sio_ver);
		case	READ_AVR_DEV:
				set_SSP_highspeed();
				if (!avr_program_enable())
					return -EIO;
				if (!avr_read_devcode(&vendor, &family, &part)) {
					avr_program_disable();
					return -EIO;
				}
				else {
					DevVer = (struct avr_ver *)arg;
					DevVer->Vendor = vendor;
					DevVer->Family = family;
					DevVer->Part = part;
					return 3*sizeof(unchar);
				}
		case	READ_FLASH_SIZE:
				(*(int *)arg) = flash_size;
				return flash_size;
		case	ERASE_CHIP:
				set_SSP_highspeed();
				if (!avr_program_enable())
					return -EIO;
				if (!send_program_msg((unchar *)&ChipErase, echo))
					return -EBUSY;
				mdelay(30);
				avr_program_disable();
				break;
		default:
			return -EINVAL;
	}

	return 1;
}

struct file_operations avr_fops = {
	open:		avr_open,
	release:	avr_release,
	read: 		avr_read,
	write:		avr_write,
	ioctl:		avr_ioctl,
};
